/*
** Module   : BaseNativeType.java
** Abstract : descriptor for a native function's argument or return value 
**
** Copyright (c) 2013-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 GES 20131211 First version.
** 002 GES 20140127 Fixed out of order memory buffer delete in restore().
*/

package com.goldencode.p2j.library;

import java.io.*;
import com.goldencode.p2j.util.*;

/**
 * Descriptor for a native function's argument or return value.
 */
public abstract class BaseNativeType
implements Externalizable
{
   /** Flag to control whether this element should be copied TO the native function. */
   protected boolean copyIn = false;
   
   /** Flag to control whether this element should be copied BACK FROM the native function. */
   protected boolean copyOut = false;
   
   /** Flag to control whether this element should be passed as a pointer. */
   protected boolean passByPointer = false;
   
   /** Data type of the instance (defined as a {@link LibraryManager} constant). */
   protected int type = -1;
   
   /** On the client side, this is used to render the data into native memory. */
   protected memptr ref = null;
   
   /** On the client side, this is used to render the data as a pointer. */
   protected memptr ptr = null;
   
   /**
    * Default constructor.
    */
   public BaseNativeType()
   {
   }
   
   /**
    * Reports whether this element should be copied TO the native function.
    *
    * @return   <code>true</code> for INPUT and INPUT-OUTPUT parameters, <code>false</code>
    *           otherwise.
    */
   public boolean isCopyIn()
   {
      return copyIn;
   }
   
   /**
    * Assigns the flag for whether this element should be copied TO the native function.
    *
    * @param    copyIn
    *           <code>true</code> for INPUT and INPUT-OUTPUT parameters, <code>false</code>
    *           otherwise.
    */
   public void setCopyIn(boolean copyIn)
   {
      this.copyIn = copyIn;
   }
   
   /**
    * Reports whether this element should be copied BACK FROM the native function.
    *
    * @return   <code>true</code> for OUTPUT and RETURN parameters, <code>false</code>
    *           otherwise.
    */
   public boolean isCopyOut()
   {
      return copyOut;
   }
   
   /**
    * Assigns the flag for whether this element should be copied BACK FROM the native function.
    *
    * @param    copyOut
    *           <code>true</code> for OUTPUT and RETURN parameters, <code>false</code>
    *           otherwise.
    */
   public void setCopyOut(boolean copyOut)
   {
      this.copyOut = copyOut;
   }
   
   /**
    * Reports whether the argument passed should be a pointer to the data element or whether the
    * data should be passed by value to the native function.
    *
    * @return   The stored value of the flag.
    */
   public boolean isPassByPointer()
   {
      return passByPointer;
   }
   
   /**
    * Assigns the flag for whether the argument passed should be a pointer to the data element or
    * whether the data should be passed by value to the native function.
    *
    * @param    passByPointer
    *           The new value of the flag.
    */
   public void setPassByPointer(boolean passByPointer)
   {
      this.passByPointer = passByPointer;
   }
   
   /**
    * Allocate native memory and render the data into that memory. If the instance requires
    * copying data back from native memory or must be forced into passing by pointer, then
    * this method will return back the address that contains the address of the memory in
    * which the data is rendered.
    * <p>
    * This should never be called on the server, only on the client side.
    *
    * @return   The memory address of the rendered data or 0 if there is no data to
    *           output.
    */
   public long render()
   {
      long addr = render(0);
      
      if (copyOut || passByPointer)
      {
         int sz = LibraryManager.ptrSize();
         
         ptr = new memptr();
         ptr.setLength(sz);
         
         if (sz == 4)
         {
            ptr.setLong(addr, 1);
         }
         else
         {
            ptr.setInt64(addr, 1);
         }
         
         addr = ptr.getPointerValue().longValue();
      }
      
      return addr;
   }
   
   /**
    * Copy back (if necessary) data from the native memory and deallocate the memory.
    * <p>
    * This should never be called on the server, only on the client side.
    */
   public void restore()
   {
      long addr = 0;
      
      if (ptr != null)
      {
         addr = ptr.getPointerValue().longValue();
      }
      
      restore(addr);
      
      if (ptr != null)
      {
         ptr.setLength(0);
         ptr = null;
      }
   }
   
   /**
    * Manually override the data type of the instance.
    *
    * @param    type
    *           A {@link LibraryManager} constant representing the data type.
    */
   public void setType(int type)
   {
      this.type = type;
   }
   
   /**
    * Report the data type of the instance.
    *
    * @return   A {@link LibraryManager} constant representing the data type.
    */
   public int type()
   {
      return (copyOut || passByPointer) ? LibraryManager.TYPE_POINTER : type;
   }
   
   /**
    * Report the size (in bytes) of the native memory needed to store this instance.
    * <p>
    * This should never be called on the server, only on the client side.
    *
    * @return   The number of bytes needed.
    */
   public abstract int size();
   
   /**
    * Render the data into native memory, which will be allocated if the address is not
    * given.
    * <p>
    * This should never be called on the server, only on the client side.
    *
    * @param    addr
    *           The memory address to which the data should be rendered, if the caller
    *           is supplying a buffer or 0 if the buffer should be allocated.
    *
    * @return   The memory address of the rendered data or 0 if there is no data to
    *           output.
    */
   public abstract long render(long addr);
   
   /**
    * Copy back (if necessary) data from the native memory and deallocate the memory.
    * <p>
    * This should never be called on the server, only on the client side.
    *
    * @param    addr
    *           The memory address from which the data should be restored, if the caller
    *           is supplying a buffer or 0 if the buffer should already exist.
    */
   public abstract void restore(long addr);
   
   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.  The parent's state is
    * read first.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      copyIn        = in.readBoolean();
      copyOut       = in.readBoolean();
      passByPointer = in.readBoolean();
      
      // the memptr is only used on the client side and it is never transferred
      // over the network
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.  The parent's state is
    * written first.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      out.writeBoolean(copyIn);
      out.writeBoolean(copyOut);
      out.writeBoolean(passByPointer);
      
      // the memptr is only used on the client side and it is never transferred
      // over the network
   }
}
